/*###########################################################################
#                                                                          	#
#                     LCDW[ SG12864A 䃂W[					#
#																			#
#  QlFgLCD Driver KS0107 128x64 for AVR									#
#																			#
###########################################################################*/
#define __SG12864A_C__

#include <avr/io.h>
#include <avr/pgmspace.h>
#include <util/delay.h>
#include "sg12864a.h"
#include "font.h"

/* RESET ݒ}N`			*/
#define RESET_ENABLE()			( PORTD &= ~_BV(PD4) )
#define RESET_DISABLE()			( PORTD |=  _BV(PD4) )

/* E ݒ}N`				*/
#define E_LOW()					( PORTB &= ~_BV(PD3) )
#define E_HI()					( PORTB |=  _BV(PD3) )

/* D/I ݒ}N`			*/
#define DI_INSTRUCTIONCODE()	( PORTB &= ~_BV(PB4) )
#define DI_DATACODE()			( PORTB |=  _BV(PB4) )

/* R/W ݒ}N`			*/
#define RW_WRITEMODE()			( PORTB &= ~_BV(PB5) )
#define RW_READMODE()			( PORTB |=  _BV(PB5) )

/* CS1 ݒ}N`			*/
#define CS1_ENABLE()			( PORTB |=  _BV(PB2) )
#define CS1_DISABLE()			( PORTB &= ~_BV(PB2) )

/* CS2 ݒ}N`			*/
#define CS2_ENABLE()			( PORTB |=  _BV(PB1) )
#define CS2_DISABLE()			( PORTB &= ~_BV(PB1) )

/* f[^|[g̓[h		*/
#define DDR_INPUT()				{ DDRC &= 0xf0;	DDRD &= 0xf0; PORTC &= 0xf0; PORTD &= 0xf0; }

/* f[^|[go̓[h		*/
#define DDR_OUTPUT()			{ DDRC |= 0x0f; DDRD |= 0x0f; PORTC &= 0xf0; PORTD &= 0xf0; }

/* e^C~O҂}N		*/
#define WAIT_TAS() _delay_us( 0.140 )	/* RW/CSm肵ĂELɂ܂ł̑҂	*/
#define WAIT_TD()  _delay_us( 0.200	)	/* ELɂĂŏ						*/
#define WAIT_TAH() _delay_us( 0.010	)	/* EȂ̎̓܂ł̍ŏ҂		*/

/* ʑI}N				*/
#define _LCD_selectLeftChip()	{ CS1_ENABLE();  CS2_DISABLE(); }	/* ʑI	*/
#define _LCD_selectRightChip()	{ CS2_ENABLE();  CS1_DISABLE(); }	/* EʑI	*/
#define _LCD_selectNothing()	{ CS1_DISABLE(); CS2_DISABLE(); }	/* I		*/

/* LCDR}hf[^			*/
#define SET_DISPLAY_OFF             	0x3e
#define SET_DISPLAY_ON              	0x3f
#define SET_COLUMN_ADDRESS( col )		( 0x40 | (col)  )			/* 0 - 63	*/
#define SET_INIT_DISPLAY_LINE( line )	( 0xc0 | (line) )			/* 0 - 63	*/
#define SET_PAGE_ADDRESS( page )		( 0xb8 | (page) )			/* 0 - 7	*/

/* Βl߂}N			*/
#define	abs(a)	( ( (a) > 0 ) ? (a) : -(a) )

/* Q̕ϐ̃f[^Xbv	*/
#define swap(a,b) { uint8_t i = a; a = b; b = i; }

/* [J֐vg^Cv錾	*/
static void DB_WRITE( uint8_t data );
static uint8_t DB_READ( void );
static void _LCD_sendCommand( uint8_t command );
static void _LCD_writeByte( uint8_t data );
static uint8_t _LCD_readByte( void );
static void _LCD_waitBusy( void );
static void _LCD_doLine( uint8_t x0 ,uint8_t y0 ,uint8_t x1 ,uint8_t y1 ,void (*pPixFun)(uint8_t, uint8_t ) );

#if LCD_USE_CIRCLE
static void _LCD_doCirc( uint8_t x1 ,uint8_t y1 ,uint8_t radius ,void (*pLineFun)(uint8_t,uint8_t,uint8_t,uint8_t) );
#endif

static uint8_t _LCD_selectChip( uint8_t x );
static void _LCD_writeByteXY( uint8_t data ,uint8_t x ,uint8_t y );
static void _LCD_copyPageTo( const prog_uint8_t *source ,uint8_t swidth ,uint8_t spage ,uint8_t sx1 ,uint8_t sx2 ,uint8_t tx1 ,uint8_t ty );
static void _LCD_copyPixelTo( const prog_uint8_t *source ,uint8_t swidth ,uint8_t sx ,uint8_t sy ,uint8_t tx ,uint8_t ty );

#if LCD_USE_RECT
static void _LCD_doRect( uint8_t x1 ,uint8_t y1 ,uint8_t x2 ,uint8_t y2 ,void (*pPixFun)(uint8_t, uint8_t ),void (*pPageFun)(uint8_t, uint8_t, uint8_t ) );
static void __LCD_onPage( uint8_t page ,uint8_t byte ,uint8_t x1 ,uint8_t x2 );
static void _LCD_invertPage( uint8_t page ,uint8_t x1 ,uint8_t x2 );
static void _LCD_offPage( uint8_t page ,uint8_t x1 ,uint8_t x2 );
static void _LCD_onPage( uint8_t page ,uint8_t x1 ,uint8_t x2 );
#endif

/*****************************************************************************
*							DB|[gf[^o
* Fuint8_t data
* ߒlFvoid
* ɍ킹ĕύXKv
*****************************************************************************/
static void DB_WRITE( uint8_t data )
{
	RW_WRITEMODE();
	DDR_OUTPUT();

	PORTC = ( PORTC & 0xf0 ) | ( data >> 4 );
	PORTD = ( PORTD & 0xf0 ) | ( data & 0x0f );

	WAIT_TAS();
	E_HI();
	WAIT_TD();
	E_LOW();
	WAIT_TAH();
}

/*****************************************************************************
*								DB|[gǂݍ
* Fvoid
* ߒlFuint8_t
* ɍ킹ĕύXKv
*****************************************************************************/
static uint8_t DB_READ( void )
{
	uint8_t data;

	RW_READMODE();
	DDR_INPUT();

	WAIT_TAS();
	E_HI();
	WAIT_TD();

	data = ( PINC << 4 ) | ( PIND & 0x0f );

	E_LOW();
	WAIT_TAH();

	return ( data );
}

/*****************************************************************************
*								LCD
* Fvoid
* ߒlFvoid
*****************************************************************************/
void LCD_init( void )
{ 
	/* Rg[|[gALL HI	*/
	RESET_DISABLE();
	E_HI();
	DI_DATACODE();
	RW_READMODE();
	CS1_ENABLE();
	CS2_ENABLE();

	/* n[hEGAZbg		*/	
	RESET_ENABLE();								/* reset the chip			*/
	for	( uint8_t i = 0 ;i < 200 ;i++ ) _delay_ms( 0.1 );
	RESET_DISABLE();
	for	( uint8_t i = 0 ;i < 250 ;i++ ) _delay_ms( 0.2 );

	/* E`bvݒ			*/
	for ( uint8_t i = 0 ;i < 2 ;i++ ) {
		if ( i == 0 ) {
			_LCD_selectLeftChip();
		} else {
			_LCD_selectRightChip();
		}
		_LCD_sendCommand( SET_INIT_DISPLAY_LINE(0) );
		_LCD_sendCommand( SET_COLUMN_ADDRESS(0) );
		_LCD_sendCommand( SET_PAGE_ADDRESS(0) );
		LCD_on();
	}

	LCD_clr();
}

/*****************************************************************************
*								ʕ\ON
* Fvoid
* ߒlFvoid
*****************************************************************************/
void LCD_on( void )
{
	_LCD_selectLeftChip();
	_LCD_sendCommand( SET_DISPLAY_ON );

	_LCD_selectRightChip();
	_LCD_sendCommand( SET_DISPLAY_ON );
}

/*****************************************************************************
*								ʕ\OFF
* Fvoid
* ߒlFvoid
*****************************************************************************/
void LCD_off( void )
{
	_LCD_selectLeftChip();
	_LCD_sendCommand( SET_DISPLAY_OFF );

	_LCD_selectRightChip();
	_LCD_sendCommand( SET_DISPLAY_OFF );
}

/*****************************************************************************
*								ʃNA
* Fvoid
* ߒlFvoid
*****************************************************************************/
void LCD_clr( void )
{
	uint8_t page;
	uint8_t colum;
	uint8_t i;

	for ( i = 0 ; i < 2 ;i++ ) {
		if ( i == 0 ) {
			_LCD_selectLeftChip();
		} else {
			_LCD_selectRightChip();
		}
		for ( page = 0 ;page < 8 ;page++ ) {
			_LCD_sendCommand( SET_COLUMN_ADDRESS( 0 ) );
			_LCD_sendCommand( SET_PAGE_ADDRESS( page ) );
			for ( colum = 0 ;colum < 64 ;colum++ ) {
				_LCD_writeByte( 0 );
			}
		}
	}
}

/*****************************************************************************
*							w̍Wփhbg\
* Fuint8_t x XW( 0 ` 127 )
*		uint8_t y YW( 0 ` 63 )
* ߒlFvoid
*****************************************************************************/
void LCD_pixelOn( uint8_t x, uint8_t y )
{
	if ( y < 64 ) {

		/* XWɉʑI		*/
		if ( x < 64 ) {
			_LCD_selectLeftChip();
		} else if ( x < 128 ) {
			_LCD_selectRightChip();
		} else {
			return;
		}

		/* VRAMAhXݒ				*/
		_LCD_sendCommand( SET_COLUMN_ADDRESS( x ) );	/* Jݒ		*/
		_LCD_sendCommand( SET_PAGE_ADDRESS( y / 8 ) );	/* y[Wݒ		*/

		/* Yʒũrbg𔽓]		*/
		uint8_t data = _LCD_readByte();
		_LCD_sendCommand( SET_COLUMN_ADDRESS( x ) );
		_LCD_writeByte( data | _BV( y % 8 ) );
	}
}

/*****************************************************************************
*							w̍W̃hbgNA
* Fuint8_t x XW( 0 ` 127 )
*		uint8_t y YW( 0 ` 63 )
* ߒlFvoid
*****************************************************************************/
void LCD_pixelOff(uint8_t x, uint8_t y)
{
	if ( y < 64 ) {

		/* XWɉʑI		*/
		if ( x < 64 ) {
			_LCD_selectLeftChip();
		} else if ( x < 128 ) {
			_LCD_selectRightChip();
		} else {
			return;
		}

		/* VRAMAhXݒ				*/
		_LCD_sendCommand( SET_COLUMN_ADDRESS( x ) );	/* Jݒ		*/
		_LCD_sendCommand( SET_PAGE_ADDRESS( y / 8 ) );	/* y[Wݒ		*/

		/* Yʒũrbg𔽓]		*/
		uint8_t data = _LCD_readByte();
		_LCD_sendCommand( SET_COLUMN_ADDRESS( x ) );
		_LCD_writeByte( data & ~_BV( y % 8 ) );
	}
}

/*****************************************************************************
*							w̍W̃hbg]
* Fuint8_t x XW( 0 ` 127 )
*		uint8_t y YW( 0 ` 63 )
* ߒlFvoid
*****************************************************************************/
void LCD_invertPixel( uint8_t x ,uint8_t y )
{
	if ( y < 64 ) {

		/* XWɉʑI		*/
		if ( x < 64 ) {
			_LCD_selectLeftChip();
		} else if ( x < 128 ) {
			_LCD_selectRightChip();
		} else {
			return;
		}

		/* VRAMAhXݒ				*/
		_LCD_sendCommand( SET_COLUMN_ADDRESS( x ) );	/* Jݒ		*/
		_LCD_sendCommand( SET_PAGE_ADDRESS( y / 8 ) );	/* y[Wݒ		*/

		/* Yʒũrbg𔽓]		*/
		uint8_t data = _LCD_readByte();
		_LCD_sendCommand( SET_COLUMN_ADDRESS( x ) );
		_LCD_writeByte( data ^ _BV( y % 8 ) );
	}
}

/*****************************************************************************
*						̃tHg(5x7)ŕ\
* Fuint8_t page y[Wԍ(0`7)
*		uint8_t x    XW(0`127)
*		uint8_t c    ASCIIR[h
* ߒlFvoid
*****************************************************************************/
void LCD_putchar( uint8_t page ,uint8_t x ,uint8_t c )
{
	const prog_uint8_t *font = font5x7[ c - 0x20 ];
	uint8_t cs;
	uint8_t i;

	/* y[WԍEJݒ		*/
	cs = _LCD_selectChip( x );
	_LCD_sendCommand( SET_COLUMN_ADDRESS( x ) );
	_LCD_sendCommand( SET_PAGE_ADDRESS( page ) );

	/* Jf[^ݒ					*/
	i = 5;
	while ( i-- && ( x < 128 ) ) {

		/* `bv`FW̏		*/
		if ( !cs && ( x >= 64 ) ) {
			cs = _LCD_selectChip( x );
			_LCD_sendCommand( SET_COLUMN_ADDRESS( x ) );
			_LCD_sendCommand( SET_PAGE_ADDRESS( page ) );
		}

		_LCD_writeByte( pgm_read_byte( font++ ) );
		x++;
	}

	if ( x < 128 ) {
		_LCD_writeByte( 0x00 );
	}
}

/*****************************************************************************
*					̃tHg(5x7)ŕ\
* Fuint8_t page y[Wԍ(0`7)
*		uint8_t x    XW(0`127)
*		char    *str 
* ߒlFvoid
*****************************************************************************/
void LCD_puts( uint8_t page ,uint8_t x ,const char *str )
{
	while ( *str ) {
		LCD_putchar( page ,x ,*str++ );
		x += 6;
	}
}

void LCD_puts_P( uint8_t page ,uint8_t x ,const prog_char *str )
{
	uint8_t c;
	while ( ( c = pgm_read_byte( str++ ) ) ) {
		LCD_putchar( page ,x ,c );
		x += 6;
	}
}

/*****************************************************************************
*				OtBbNWɓtHg(5x7)ŕ`
* Fuint8_t x XW
*		uint8_t y YW
*		char    c R[h
* ߒlFvoid
*****************************************************************************/
void LCD_Gputchar( uint8_t x ,uint8_t y ,char c )
{
	const prog_uint8_t *font = font5x7[ c - 0x20 ];
	LCD_drawBMP( font ,x ,y ,5 ,7 );
}

/*****************************************************************************
*					̃tHg(5x7)ŕ\
* Fuint8_t x    XW(0`127)
*		uint8_t y    YW(0`63)
*		char    *str 
* ߒlFvoid
*****************************************************************************/
void LCD_Gputs( uint8_t x ,uint8_t y ,const char *str )
{
	while ( *str ) {
		LCD_Gputchar( x ,y ,*str++ );
		x += 6;
	}
}

void LCD_Gputs_P( uint8_t x ,uint8_t y ,const prog_char *str )
{
	uint8_t c;
	while ( ( c = pgm_read_byte( str++ ) ) ) {
		LCD_Gputchar( x ,y ,c );
		x += 6;
	}
}

/*****************************************************************************
*								`ihbgZbgj
* Fuint8_t x0 n_XW
*		uint8_t y0 n_YW
*		uint8_t x1 I_XW
*		uint8_t y1 I_YW
* ߒlFvoid
*****************************************************************************/
void LCD_drawLine( uint8_t x0 ,uint8_t y0 ,uint8_t x1 ,uint8_t y1 )
{
	_LCD_doLine( x0 ,y0 ,x1 ,y1 ,&LCD_pixelOn );
}

/*****************************************************************************
*								`(hbgNA)
* Fuint8_t x0 n_XW
*		uint8_t y0 n_YW
*		uint8_t x1 I_XW
*		uint8_t y1 I_YW
* ߒlFvoid
*****************************************************************************/
void LCD_clearLine(uint8_t x0, uint8_t y0, uint8_t x1, uint8_t y1)
{
	_LCD_doLine( x0 ,y0 ,x1 ,y1 ,&LCD_pixelOff );
}

/*****************************************************************************
*								`(hbg])
* Fuint8_t x0 n_XW
*		uint8_t y0 n_YW
*		uint8_t x1 I_XW
*		uint8_t y1 I_YW
* ߒlFvoid
*****************************************************************************/
void LCD_invertLine( uint8_t x0 ,uint8_t y0 ,uint8_t x1 ,uint8_t y1 )
{
	_LCD_doLine( x0 ,y0 ,x1 ,y1 ,&LCD_invertPixel );
}

/*****************************************************************************
*								``iĝ݃Zbg)
* Fuint8_t x1 XW
*		uint8_t y1 YW
*		uint8_t x2 EXW
*		uint8_t y2 EYW
* ߒlFvoid
*****************************************************************************/
void LCD_drawRect( uint8_t x1 ,uint8_t y1 ,uint8_t x2 ,uint8_t y2 )
{
	LCD_drawLine( x1 ,y1 ,x2 ,y1 );
	LCD_drawLine( x1 ,y1 ,x1 ,y2 );
	LCD_drawLine( x2 ,y2 ,x2 ,y1 );
	LCD_drawLine( x1 ,y2 ,x2 ,y2 );  
}

#if LCD_USE_RECT
/*****************************************************************************
*								``ihbgZbgj
* Fuint8_t x1 XW
*		uint8_t y1 YW
*		uint8_t x2 EXW
*		uint8_t y2 EYW
* ߒlFvoid
*****************************************************************************/
void LCD_fillRect( uint8_t x1 ,uint8_t y1 ,uint8_t x2 ,uint8_t y2 )
{
	_LCD_doRect( x1 ,y1 ,x2 ,y2 ,&LCD_pixelOn ,&_LCD_onPage );
}

/*****************************************************************************
*								``ihbgNAj
* Fuint8_t x1 XW
*		uint8_t y1 YW
*		uint8_t x2 EXW
*		uint8_t y2 EYW
* ߒlFvoid
*****************************************************************************/
void LCD_clearRect( uint8_t x1 ,uint8_t y1 ,uint8_t x2 ,uint8_t y2 )
{
	_LCD_doRect( x1 ,y1 ,x2 ,y2 ,&LCD_pixelOff ,&_LCD_offPage );
}

/*****************************************************************************
*								``ihbg]j
* Fuint8_t x1 XW
*		uint8_t y1 YW
*		uint8_t x2 EXW
*		uint8_t y2 EYW
* ߒlFvoid
*****************************************************************************/
void LCD_invertRect( uint8_t x1 ,uint8_t y1 ,uint8_t x2 ,uint8_t y2 )
{
	_LCD_doRect( x1 ,y1 ,x2 ,y2 , &LCD_invertPixel ,&_LCD_invertPage );
}
#endif /* LCD_USE_RECT */

#if LCD_USE_CIRCLE
/*****************************************************************************
*						~`(Oĝ)
* Fuint8_t x1     SXW(0`127)
*		uint8_t y1     SYW(0`63)
*		uint8_t radius a
* ߒlFvoid
*****************************************************************************/
void LCD_drawCirc( uint8_t x1 ,uint8_t y1 ,uint8_t radius )
{
	uint8_t y = 0;
	uint8_t x = 0;
	uint8_t d = 0;
	int8_t part;

	if ( radius ) {
		d = y1 - x1;
		y = radius;
		part = 3 - 2 * radius;

		while ( x <= y ) {
			LCD_pixelOn( x1 + x    , y1 + y );
			LCD_pixelOn( x1 + x    , y1 - y );
			LCD_pixelOn( x1 - x    , y1 + y );
			LCD_pixelOn( x1 - x    , y1 - y );
			LCD_pixelOn( y1 + y - d, y1 + x );
			LCD_pixelOn( y1 + y - d, y1 - x );
			LCD_pixelOn( y1 - y - d, y1 + x );
			LCD_pixelOn( y1 - y - d, y1 - x );
			if ( part < 0 ) {
				part += ( 4 * x + 6 );
			} else {
				part += ( 4 * ( x - y ) + 10 );
				y--;
			}
			x++;
		}
	}
}

/*****************************************************************************
*							hԂ~`(Zbg)
* Fuint8_t x1     SXW(0`127)
*		uint8_t y1     SYW(0`63)
*		uint8_t radius a
* ߒlFvoid
*****************************************************************************/
void LCD_fillCirc(uint8_t x1, uint8_t y1, uint8_t radius)
{  
	_LCD_doCirc( x1 ,y1 ,radius ,&LCD_drawLine );
}

/*****************************************************************************
*							hԂ~`(NA)
* Fuint8_t x1     SXW(0`127)
*		uint8_t y1     SYW(0`63)
*		uint8_t radius a
* ߒlFvoid
*****************************************************************************/
void LCD_clearCirc(uint8_t x1, uint8_t y1, uint8_t radius)
{  
	_LCD_doCirc( x1 ,y1 ,radius ,&LCD_clearLine );
}

/*****************************************************************************
*							hԂ~`(])
* Fuint8_t x1     SXW(0`127)
*		uint8_t y1     SYW(0`63)
*		uint8_t radius a
* ߒlFvoid
*****************************************************************************/
void LCD_invertCirc(uint8_t x1, uint8_t y1, uint8_t radius)
{  
	_LCD_doCirc( x1 ,y1 ,radius ,&LCD_invertLine );
}
#endif /* LCD_USE_CIRCLE	*/

/*****************************************************************************
*							128x64dott摜\
* Fprog_uint8_t *bitmap C[W(1024bytes)
* ߒlFvoid
*****************************************************************************/
void LCD_drawFullBMP( const prog_uint8_t *bitmap )
{  
	uint8_t page ,x;
	for ( page = 0 ;page < 8 ;page++ ) {
		_LCD_selectLeftChip();
		_LCD_sendCommand( SET_COLUMN_ADDRESS(0) );
		_LCD_sendCommand( SET_PAGE_ADDRESS(page) );
		for ( x = 0 ; x < 64 ;x++ ) {
			_LCD_writeByte( pgm_read_byte( bitmap++ ) );
	    }

		_LCD_selectRightChip();
		_LCD_sendCommand( SET_COLUMN_ADDRESS(0) );
		_LCD_sendCommand( SET_PAGE_ADDRESS(page) );
		for ( x = 0 ; x < 64 ;x++ ) {
			_LCD_writeByte( pgm_read_byte( bitmap++ ) );
	    }
	}
}

/*****************************************************************************
*							w̉摜\
* Fprog_uint8_t *bitmap C[W(1024bytes)
*		uint8_t x            \JnXW
*		uint8_t y			 \JnYW
*		uint8_t width		 C[W̉hbg
*		uint8_t height       C[W̏chbg
* ߒlFvoid
*****************************************************************************/
void LCD_drawBMP( const prog_uint8_t *bitmap ,uint8_t x ,uint8_t y ,uint8_t width ,uint8_t height )
{
	uint8_t pages ,i ,j;

	/* whole pages  	*/
	pages = height / 8;
	j = 0;

	while( j < pages ) {
		_LCD_copyPageTo( bitmap ,width ,j ,0 ,width ,x ,y + ( j * 8 ) );
		j++;
	}

	/* single lines left      					*/
	/*  single_lines=height%LCD_PAGE_HEIGHT;	*/
	j *= 8;
	while ( j < height ) {
		for( i = 0 ; i < width ;i++ ) {
			_LCD_copyPixelTo( bitmap ,width ,i ,j ,x + i ,y + j );  
		}
		j++;
	}
}

/****************************************************************************/
/****************************************************************************/
/******************** ȉ[J֐ **************************************/
/****************************************************************************/
/****************************************************************************/

/*****************************************************************************
*							LCDR}ho
* Fuint8_t command R}h
* ߒlFvoid
*****************************************************************************/
static void _LCD_sendCommand( uint8_t command )
{
	_LCD_waitBusy();
	DI_INSTRUCTIONCODE();
	DB_WRITE( command );
}

/*****************************************************************************
*								LCDf[^o
* Fuint8_t data f[^
* ߒlFvoid
*****************************************************************************/
static void _LCD_writeByte( uint8_t data )
{
	_LCD_waitBusy();
	DI_DATACODE();
	DB_WRITE( data );
}

/*****************************************************************************
*								LCDf[^
* Fvoid
* ߒlFuint8_t f[^
*****************************************************************************/
static uint8_t _LCD_readByte( void )
{
	_LCD_waitBusy();
	DI_DATACODE();
	DB_READ();

	_LCD_waitBusy();
	DI_DATACODE();
	return ( DB_READ() );
}

/*****************************************************************************
*							LCDbusẙԑ҂
*@Fvoid
*@ߒlFvoid
*****************************************************************************/
static void _LCD_waitBusy( void )
{
	DI_INSTRUCTIONCODE();			/* D/Iݒ (Low:Instruction mode)		*/
	while ( DB_READ() & 0x80 );		/* BusytOL̊ԃ[v			*/
}

/*****************************************************************************
*						`Tu֐
* Fuint8_t x0 n_XW
*		uint8_t y0 n_YW
*		uint8_t x1 I_XW
*		uint8_t y1 I_YW
*		void (*pPixFun)(uint8_t, uint8_t )) hbg`֐ւ̃|C^
* ߒlFvoid
****************************************************************************/
static void _LCD_doLine( uint8_t x0 ,uint8_t y0 ,uint8_t x1 ,uint8_t y1 ,void (*pPixFun)(uint8_t, uint8_t ) )
{
	int steep;
	int deltax;
	int deltay;
	int error;
	uint8_t x;
	uint8_t y;
	int ystep;

	/* ̑傫ق߂				*/
	steep = ( abs( y1 - y0 ) > abs( x1 - x0 ) );

	/* X ,Y ̓ւ						*/
	if ( steep ) {
		swap( x0 ,y0 );
		swap( x1 ,y1 );
	}

	if( x0 > x1 ) {
		swap( x0 ,x1 );
		swap( y0 ,y1 );
	}

	deltax = (int8_t)x1 - (int8_t)x0;			/* XvZ		*/
	deltay = abs( y1 - y0 );
	error = 0;
	y = y0;

	/* XŃXebv̐؂ւ			*/
	if( y0 < y1 ) {
		ystep = 1;
	} else {
		ystep = -1;
	}

	/* _ŕ`							*/
	for ( x = x0 ;x <= x1 ;x++ ) {
		if ( steep ) {
			pPixFun( y, x );
		} else {
			pPixFun( x, y );
		}
		error += deltay;
		if ( ( error << 1 ) >= deltax ) {
			y += ystep;
			error -= deltax;
		}
	}
}

#if LCD_USE_CIRCLE
/*****************************************************************************
*						~`Tu
* Fuint8_t x1     SXW(0`127)
*		uint8_t y1     SYW(0`63)
*		uint8_t radius a
*       void (*pLineFun)(uint8_t,uint8_t,uint8_t,uint8_t) C֐
* ߒlFvoid
*****************************************************************************/
static void _LCD_doCirc( uint8_t x1 ,uint8_t y1 ,uint8_t radius ,void (*pLineFun)(uint8_t,uint8_t,uint8_t,uint8_t) )
{
	uint8_t y = 0;
	uint8_t x = 0;
	uint8_t d = 0;
	int8_t part;

	d = y1 - x1;
	y = radius;
	part = 3 - 2 * radius;

	while ( x <= y ) {
		pLineFun( x1 + x     ,y1 + y ,x1 + x     ,y1 - y );
		pLineFun( x1 - x     ,y1 + y ,x1 - x     ,y1 - y );
		pLineFun( y1 + y - d ,y1 + x ,y1 + y - d ,y1 - x );
		pLineFun( y1 - y - d ,y1 + x ,y1 - y - d ,y1 - x );
		if ( part < 0 ) {
			part += ( ( 4 * x ) + 6 );
		} else {
			part += ( ( 4 * ( x - y ) ) + 10 );
			y--;
		}
		x++;
	}
}
#endif /* LCD_USE_CIRCLE */

/*****************************************************************************
*						XWɉ`bvI
* Fuint8_t x XW
* ߒlFuint8_t   I`bvԍ( 0 or 1 )
*****************************************************************************/
static uint8_t _LCD_selectChip( uint8_t x )
{
	uint8_t cs;

	if ( x < 64 ) {
		_LCD_selectLeftChip();
		cs = 0;
	} else {
		_LCD_selectRightChip();
		cs = 1;
	}
	return ( cs );
}

/*****************************************************************************
*					wWփoCg
* Fuint8_t data oCgf[^(cɂWhbg`j
*       uint8_t x
*		uint8_t y
* ߒlFvoid
*****************************************************************************/
static void _LCD_writeByteXY( uint8_t data ,uint8_t x ,uint8_t y )
{
	uint8_t i = 8;
	if ( x < 128 ) {
		while ( i-- && ( y < 64 ) ) {
			if ( data & 0x01 ) {
				LCD_pixelOn( x ,y++ );
			} else {
				LCD_pixelOff( x ,y++ );
			}
			data >>= 1;
		}
	}
}

/*****************************************************************************
*					y[WPʂł̃C[W`
* Fconst prog_uint8_t *source
*		uint8_t swidth
*		uint8_t spage
*		uint8_t sx1
*		uint8_t sx2
*		uint8_t tx
*		uint8_t ty
* ߒlFvoid
*****************************************************************************/
static void _LCD_copyPageTo( const prog_uint8_t *source ,uint8_t swidth ,uint8_t spage ,uint8_t sx1 ,uint8_t sx2 ,uint8_t tx1 ,uint8_t ty )
{
	uint8_t byte ,i ,tx2;  
	if ( ( sx2 + tx1 ) > 128 ) {
		tx2 = 127;
	} else {
		tx2 = tx1 + ( sx2 - sx1 );    
	}

	for ( i = 0 ; i < swidth ;i++ ) {
		byte = pgm_read_byte( &source[ spage * swidth + i ] );  
		if( ( ( tx1 + i ) < 128 ) && ( ty < 64 ) ) {
			_LCD_writeByteXY( byte , i + tx1 ,ty );
		}
	}
}

/*****************************************************************************
*					hbgPʂł̃C[W`
* Fconst prog_uint8_t *source
*		uint8_t swidth
*		uint8_t spage
*		uint8_t sx1
*		uint8_t sx2
*		uint8_t tx
*		uint8_t ty
* ߒlFvoid
*****************************************************************************/
static void _LCD_copyPixelTo( const prog_uint8_t *source ,uint8_t swidth ,uint8_t sx ,uint8_t sy ,uint8_t tx ,uint8_t ty )
{
	uint8_t byte;
	if ( ( tx < 128 ) && ( ty < 64 ) ) {
		byte = pgm_read_byte( &source[ ( sy / 8 ) * swidth + sx ] );
		if ( byte & ( 1 << ( sy % 8 ) ) ) {
			LCD_pixelOn( tx ,ty );
		} else {
			LCD_pixelOff( tx ,ty );
		}
	}
}

#if LCD_USE_RECT
/*****************************************************************************
*								``Tu֐
* Fuint8_t x1 XW
*		uint8_t y1 YW
*		uint8_t x2 EXW
*		uint8_t y2 EYW
*		void (*pPixFun)(uint8_t, uint8_t )) hbg`֐ւ̃|C^
*		void (*pPageFun)(uint8_t, uint8_t , uint8_t )) y[WhԂ֐ւ̃|C^
* ߒlFvoid
****************************************************************************/
static void _LCD_doRect( uint8_t x1 ,uint8_t y1 ,uint8_t x2 ,uint8_t y2 ,void (*pPixFun)(uint8_t, uint8_t ),void (*pPageFun)(uint8_t, uint8_t, uint8_t ) )
{
	uint8_t lines ,x;

	if ( y1 > y2 ) {				/* YWʒuƂ				*/
		swap( y1 ,y2 );
	}

	if ( x1 > x2 ) {				/* XWʒuƂ					*/
		swap( x1 ,x2 );
	}

	/* oCgE܂ŃhbgPʂŃC`	*/
	lines = 8 - ( y1 % 8 );
	if ( lines != 8 ) {
		while ( lines-- && ( y1 <= y2 ) ) {
			for( x = x1 ;x <= x2 ;x++ ) {
				pPixFun( x ,y1 );
			}
			y1++;
		}
	}

	/* y[WPʂŕ`						*/
	while ( ( y1 <= y2 ) && ( ( y2 - y1 ) >= 8 ) ) {
		pPageFun( y1 / 8 ,x1 ,x2 );
		y1 += 8;
	}

	/* c̃ChbgPʂŕ`		*/
	while ( y1 <= y2 ) {
		for( x = x1 ;x <= x2 ;x++ ) {
			pPixFun( x, y1 );
		}
		y1++;
	}
}

/*****************************************************************************
*					w̃y[W̃J͈͂wf[^œhԂ
* Fuint8_t page y[Wԍ( 0 ` 7 )
*		uint8_t byte Zbgf[^
*		uint8_t x1   JnXW( 0 ` 127 )
*		uint8_t y2   IXW( 0 ` 127 )
* ߒlFvoid
*****************************************************************************/
static void __LCD_onPage( uint8_t page ,uint8_t byte ,uint8_t x1 ,uint8_t x2 )
{
	uint8_t x;
	uint8_t cs;

	if ( ( x1 < 128 ) && ( x2 < 128 ) && ( page < 8 ) ) {
		if ( x1 > x2 ) {
			swap( x1 ,x2 );
		}

		/* y[WԍEJݒ		*/
		cs = _LCD_selectChip( x1 );
		_LCD_sendCommand( SET_PAGE_ADDRESS( page ) );
		_LCD_sendCommand( SET_COLUMN_ADDRESS( x1 ) );

		/* Jf[^ݒ					*/
		for ( x = x1 ; x <= x2 ;x++ ) {

			/* `bv`FW̏		*/
			if ( !cs && ( x >= 64 ) ) {
				cs = 1;
				_LCD_selectRightChip();
				_LCD_sendCommand( SET_PAGE_ADDRESS( page ) );
				_LCD_sendCommand( SET_COLUMN_ADDRESS( x ) );
			}

			/* oCgf[^Zbg			*/
			_LCD_writeByte( byte );
		}
	}
}

/*****************************************************************************
*							w̃y[W̃J͈͂𔽓]
* Fuint8_t page y[Wԍ( 0 ` 7 )
*		uint8_t x1   JnXW( 0 ` 127 )
*		uint8_t y2   IXW( 0 ` 127 )
* ߒlFvoid
* y[WԍYWvZꍇ Y = page * 8
*****************************************************************************/
static void _LCD_invertPage( uint8_t page ,uint8_t x1 ,uint8_t x2 )
{
	uint8_t mask;
	uint8_t x;
	uint8_t cs;

	if ( ( x1 < 128 ) && ( x2 < 128 ) && ( page < 8 ) ) {

		/* XWl̑召֌W␳			*/
		if ( x1 > x2 ) {
			swap( x1 ,x2 );
		}

		/* y[Wԍݒ					*/
		cs = _LCD_selectChip( x1 );
		_LCD_sendCommand( SET_PAGE_ADDRESS( page ) );

		/* Jf[^ݒ					*/
		for ( x = x1 ;x <= x2 ;x++ ) {

			/* `bv`FW̏		*/
			if ( !cs && ( x >= 64 ) ) {
				cs = 1;
				_LCD_selectRightChip();
				_LCD_sendCommand( SET_PAGE_ADDRESS( page ) );
			}

			/* YoCgf[^擾			*/
			_LCD_sendCommand( SET_COLUMN_ADDRESS( x ) );
			mask = _LCD_readByte();

			/* YoCgf[^]			*/
			_LCD_sendCommand( SET_COLUMN_ADDRESS( x ) );
			_LCD_writeByte( ~mask );
		}
	}
}

/*****************************************************************************
*					w̃y[W̃J͈͂NA
* Fuint8_t page y[Wԍ( 0 ` 7 )
*		uint8_t x1   JnXW( 0 ` 127 )
*		uint8_t y2   IXW( 0 ` 127 )
* ߒlFvoid
* y[WԍYWvZꍇ Y = page * 8
*****************************************************************************/
static void _LCD_offPage( uint8_t page ,uint8_t x1 ,uint8_t x2 )
{
	__LCD_onPage( page ,0x00 ,x1 ,x2 );
}

/*****************************************************************************
*					w̃y[W̃J͈͂Zbg
* Fuint8_t page y[Wԍ( 0 ` 7 )
*		uint8_t x1   JnXW( 0 ` 127 )
*		uint8_t y2   IXW( 0 ` 127 )
* ߒlFvoid
* y[WԍYWvZꍇ Y = page * 8
*****************************************************************************/
static void _LCD_onPage( uint8_t page ,uint8_t x1 ,uint8_t x2 )
{
	__LCD_onPage( page ,0xff ,x1 ,x2 );
}
#endif /* LCD_USE_RECT */
